package model;

import model.elements.Ant;
import model.elements.Cell;
import model.elements.Direction;
import model.types.CellType;
import model.types.DomainType;
import model.factory.MazeFactory;
import model.types.Point;
import model.factory.PopulationFactory;

import java.nio.file.Paths;
import java.util.LinkedList;
import java.util.List;

/**
 * Created by filip on 08.04.16..
 */
public class World {

    public static Integer NUMBER_OF_AGENTS    = 1000;
    public static Integer NUMBER_OF_SOLUTIONS = 10;
    public static Integer BEST_SOLUTIONS      = 10;
    public static Double  ALPHA               = 2.0;
    public static Double  EVAPORATION_RATE    = 0.8;


    public static DomainType DEFAULT_DOMAIN_TYPE = DomainType.MEDIUM_NORMAL;

    private List<Ant> population;
    private Cell[][] domain;
    private Point nest;
    private Point food;

    public World() {
        this(DEFAULT_DOMAIN_TYPE, NUMBER_OF_AGENTS, NUMBER_OF_SOLUTIONS);
    }

    public Point getNest(){
        return nest;
    }
    public World(DomainType type, int numberOfAgents, int numberOfFoodSpots) {
        generateNewDomain(type, numberOfAgents, numberOfFoodSpots);
    }

    public void generateNewDomain(DomainType type, int numberOfAgents, int numberOfFoodSpots) {
        nest = new Point(0,0);
        food = new Point(0,0);
        domain = MazeFactory.generateSimpleMaze(type, nest, food);
    //    domain = MazeFactory.parseMazeFromFile(Paths.get("/home/filip/maze"), food, nest);
     //   domain = MazeFactory.schoolExampleMaze(type, nest, food);
        population = PopulationFactory.generatePopulation(numberOfAgents, nest);
        domain[nest.getY()][nest.getX()].setNest();
    }

    public int getHeight() {
        return domain.length;
    }

    public int getWidth() {
        return domain[0].length;
    }

    public Cell getElement(int x, int y) {
        return domain[y][x];
    }

    public boolean isLegalPosition(Point p){
        return p.getX()>=0 &&
               p.getX()<getWidth() &&
               p.getY()>=0 &&
               p.getY()<getHeight() &&
               getElement(p).getType() != CellType.WALL;
    }

    public Cell getElement(Point p){
        return domain[p.getY()][p.getX()];
    }

    public Point getFood(){
        return food;
    }

    public List<Ant> getPopulation() {
        return population;
    }

    public void resetPopulaton(){
        population = PopulationFactory.generatePopulation(population.size(), nest);
    }

    public void resetPopulation(int n){
        population = PopulationFactory.generatePopulation(n, nest);
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        for(int j=0;j<getHeight();++j){
            for(int i=0;i<getWidth();++i){
                sb.append(domain[j][i]).append(" ");
            }
            sb.append("\n");
        }
        return sb.toString();
    }

    public List<Point> getNeighbours(Point position) {
        List<Point> neighbours = new LinkedList<>();
        Point pos;

        pos = position.add(Direction.NORTH.getDirectionVector());
        if(isLegalPosition(pos)) neighbours.add(pos);

        pos = position.add(Direction.EAST.getDirectionVector());
        if(isLegalPosition(pos)) neighbours.add(pos);

        pos = position.add(Direction.SOUTH.getDirectionVector());
        if(isLegalPosition(pos)) neighbours.add(pos);

        pos = position.add(Direction.WEST.getDirectionVector());
        if(isLegalPosition(pos)) neighbours.add(pos);

        return neighbours;
    }

    public void evaporatePheromones() {
        for(int i = 0; i < domain.length; ++i){
            for(int j=0; j< domain[i].length; ++j){
                domain[i][j].decreasePheromon();
            }
        }
    }

    public void setPopulation(List<Ant> population){
        this.population = population;
    }

    public void reMake(Cell[][] maze, Point food, Point nest){
        this.food = food;
        this.nest = nest;
        this.domain = maze;
        population = PopulationFactory.generatePopulation(NUMBER_OF_AGENTS, nest);
        Cell.pherMax = 0.19;
    }

    public void reset() {
        for(int i = 0; i < domain.length; ++i){
            for(int j=0; j< domain[i].length; ++j){
                domain[i][j].setPheromoneLevel(1E-3);
            }
        }
        Cell.pherMax = 0.1;
        resetPopulaton();
    }

    public void setMaze(Cell[][] maze) {
        this.domain = maze;
    }
}
